import { HttpService } from '@nestjs/axios';
import { IGeorouter } from '../../domain/interfaces/georouter.interface';
import { GeorouterSettings } from '../../domain/types/georouter-settings.type';
import { Path } from '../../domain/types/path.type';
import { Injectable } from '@nestjs/common';
import { catchError, lastValueFrom, map } from 'rxjs';
import { AxiosError, AxiosResponse } from 'axios';
import { IGeodesic } from '../../domain/interfaces/geodesic.interface';
import { NamedRoute } from '../../domain/entities/ecosystem/named-route';
import { Route } from '../../domain/entities/ecosystem/route';
import { SpacetimePoint } from '../../domain/entities/ecosystem/spacetime-point';
import {
  MatcherException,
  MatcherExceptionCode,
} from '../../exceptions/matcher.exception';

@Injectable()
export class GraphhopperGeorouter implements IGeorouter {
  _url: string;
  _urlArgs: Array<string>;
  _withTime: boolean;
  _withPoints: boolean;
  _withDistance: boolean;
  _paths: Array<Path>;
  _httpService: HttpService;
  _geodesic: IGeodesic;

  constructor(url: string, httpService: HttpService, geodesic: IGeodesic) {
    this._url = url + '/route?';
    this._httpService = httpService;
    this._geodesic = geodesic;
  }

  route = async (
    paths: Array<Path>,
    settings: GeorouterSettings,
  ): Promise<Array<NamedRoute>> => {
    this._setDefaultUrlArgs();
    this._setWithTime(settings.withTime);
    this._setWithPoints(settings.withPoints);
    this._setWithDistance(settings.withDistance);
    this._paths = paths;
    return await this._getRoutes();
  };

  _setDefaultUrlArgs = (): void => {
    this._urlArgs = [
      'vehicle=car',
      'weighting=fastest',
      'points_encoded=false',
    ];
  };

  _setWithTime = (withTime: boolean): void => {
    this._withTime = withTime;
    if (withTime) {
      this._urlArgs.push('details=time');
    }
  };

  _setWithPoints = (withPoints: boolean): void => {
    this._withPoints = withPoints;
    if (!withPoints) {
      this._urlArgs.push('calc_points=false');
    }
  };

  _setWithDistance = (withDistance: boolean): void => {
    this._withDistance = withDistance;
    if (withDistance) {
      this._urlArgs.push('instructions=true');
    } else {
      this._urlArgs.push('instructions=false');
    }
  };

  _getRoutes = async (): Promise<Array<NamedRoute>> => {
    const routes = Promise.all(
      this._paths.map(async (path) => {
        const url: string = [
          this._getUrl(),
          '&point=',
          path.points
            .map((point) => [point.lat, point.lon].join())
            .join('&point='),
        ].join('');
        const route = await lastValueFrom(
          this._httpService.get(url).pipe(
            map((res) => (res.data ? this._createRoute(res) : undefined)),
            catchError((error: AxiosError) => {
              throw new MatcherException(
                MatcherExceptionCode.INTERNAL,
                'Georouter unavailable : ' + error.message,
              );
            }),
          ),
        );
        return <NamedRoute>{
          key: path.key,
          route,
        };
      }),
    );
    return routes;
  };

  _getUrl = (): string => {
    return [this._url, this._urlArgs.join('&')].join('');
  };

  _createRoute = (response: AxiosResponse<GraphhopperResponse>): Route => {
    const route = new Route(this._geodesic);
    if (response.data.paths && response.data.paths[0]) {
      const shortestPath = response.data.paths[0];
      route.distance = shortestPath.distance ?? 0;
      route.duration = shortestPath.time ? shortestPath.time / 1000 : 0;
      if (shortestPath.points && shortestPath.points.coordinates) {
        route.setPoints(
          shortestPath.points.coordinates.map((coordinate) => ({
            lon: coordinate[0],
            lat: coordinate[1],
          })),
        );
        if (
          shortestPath.details &&
          shortestPath.details.time &&
          shortestPath.snapped_waypoints &&
          shortestPath.snapped_waypoints.coordinates
        ) {
          let instructions: Array<GraphhopperInstruction> = [];
          if (shortestPath.instructions)
            instructions = shortestPath.instructions;
          route.setSpacetimePoints(
            this._generateSpacetimePoints(
              shortestPath.points.coordinates,
              shortestPath.snapped_waypoints.coordinates,
              shortestPath.details.time,
              instructions,
            ),
          );
        }
      }
    }
    return route;
  };

  _generateSpacetimePoints = (
    points: Array<Array<number>>,
    snappedWaypoints: Array<Array<number>>,
    durations: Array<Array<number>>,
    instructions: Array<GraphhopperInstruction>,
  ): Array<SpacetimePoint> => {
    const indices = this._getIndices(points, snappedWaypoints);
    const times = this._getTimes(durations, indices);
    const distances = this._getDistances(instructions, indices);
    return indices.map(
      (index) =>
        new SpacetimePoint(
          points[index],
          times.find((time) => time.index == index)?.duration,
          distances.find((distance) => distance.index == index)?.distance,
        ),
    );
  };

  _getIndices = (
    points: Array<Array<number>>,
    snappedWaypoints: Array<Array<number>>,
  ): Array<number> => {
    const indices = snappedWaypoints.map((waypoint) =>
      points.findIndex(
        (point) => point[0] == waypoint[0] && point[1] == waypoint[1],
      ),
    );
    if (indices.find((index) => index == -1) === undefined) return indices;
    const missedWaypoints = indices
      .map(
        (value, index) =>
          <
            {
              index: number;
              originIndex: number;
              waypoint: Array<number>;
              nearest: number;
              distance: number;
            }
          >{
            index: value,
            originIndex: index,
            waypoint: snappedWaypoints[index],
            nearest: undefined,
            distance: 999999999,
          },
      )
      .filter((element) => element.index == -1);
    for (const index in points) {
      for (const missedWaypoint of missedWaypoints) {
        const inverse = this._geodesic.inverse(
          missedWaypoint.waypoint[0],
          missedWaypoint.waypoint[1],
          points[index][0],
          points[index][1],
        );
        if (inverse.distance < missedWaypoint.distance) {
          missedWaypoint.distance = inverse.distance;
          missedWaypoint.nearest = parseInt(index);
        }
      }
    }
    for (const missedWaypoint of missedWaypoints) {
      indices[missedWaypoint.originIndex] = missedWaypoint.nearest;
    }
    return indices;
  };

  _getTimes = (
    durations: Array<Array<number>>,
    indices: Array<number>,
  ): Array<{ index: number; duration: number }> => {
    const times: Array<{ index: number; duration: number }> = [];
    let duration = 0;
    for (const [origin, destination, stepDuration] of durations) {
      let indexFound = false;
      const indexAsOrigin = indices.find((index) => index == origin);
      if (
        indexAsOrigin !== undefined &&
        times.find((time) => origin == time.index) == undefined
      ) {
        times.push({
          index: indexAsOrigin,
          duration: Math.round(stepDuration / 1000),
        });
        indexFound = true;
      }
      if (!indexFound) {
        const indexAsDestination = indices.find(
          (index) => index == destination,
        );
        if (
          indexAsDestination !== undefined &&
          times.find((time) => destination == time.index) == undefined
        ) {
          times.push({
            index: indexAsDestination,
            duration: Math.round((duration + stepDuration) / 1000),
          });
          indexFound = true;
        }
      }
      if (!indexFound) {
        const indexInBetween = indices.find(
          (index) => origin < index && index < destination,
        );
        if (indexInBetween !== undefined) {
          times.push({
            index: indexInBetween,
            duration: Math.round((duration + stepDuration / 2) / 1000),
          });
        }
      }
      duration += stepDuration;
    }
    return times;
  };

  _getDistances = (
    instructions: Array<GraphhopperInstruction>,
    indices: Array<number>,
  ): Array<{ index: number; distance: number }> => {
    let distance = 0;
    const distances: Array<{ index: number; distance: number }> = [
      {
        index: 0,
        distance,
      },
    ];
    for (const instruction of instructions) {
      distance += instruction.distance;
      if (
        (instruction.sign == GraphhopperSign.SIGN_WAYPOINT ||
          instruction.sign == GraphhopperSign.SIGN_FINISH) &&
        indices.find((index) => index == instruction.interval[0]) !== undefined
      ) {
        distances.push({
          index: instruction.interval[0],
          distance: Math.round(distance),
        });
      }
    }
    return distances;
  };
}

type GraphhopperResponse = {
  paths: [
    {
      distance: number;
      weight: number;
      time: number;
      points_encoded: boolean;
      bbox: Array<number>;
      points: GraphhopperCoordinates;
      snapped_waypoints: GraphhopperCoordinates;
      details: {
        time: Array<Array<number>>;
      };
      instructions: Array<GraphhopperInstruction>;
    },
  ];
};

type GraphhopperCoordinates = {
  coordinates: Array<Array<number>>;
};

type GraphhopperInstruction = {
  distance: number;
  heading: number;
  sign: GraphhopperSign;
  interval: Array<number>;
  text: string;
};

enum GraphhopperSign {
  SIGN_START = 0,
  SIGN_FINISH = 4,
  SIGN_WAYPOINT = 5,
}
