import {
  MatcherException,
  MatcherExceptionCode,
} from '../../../exceptions/matcher.exception';
import { IRequestGeography } from '../../interfaces/geography-request.interface';
import { PointType } from '../../types/geography.enum';
import { Point } from '../../types/point.type';
import { find } from 'geo-tz';
import { Route } from './route';
import { Role } from '../../types/role.enum';
import { IGeorouter } from '../../interfaces/georouter.interface';
import { Waypoint } from './waypoint';
import { Actor } from './actor';
import { Person } from './person';
import { Step } from '../../types/step.enum';
import { Path } from '../../types/path.type';

export class Geography {
  _geographyRequest: IRequestGeography;
  _person: Person;
  _points: Array<Point>;
  originType: PointType;
  destinationType: PointType;
  timezones: Array<string>;
  driverRoute: Route;
  passengerRoute: Route;

  constructor(
    geographyRequest: IRequestGeography,
    defaultTimezone: string,
    person: Person,
  ) {
    this._geographyRequest = geographyRequest;
    this._person = person;
    this._points = [];
    this.originType = undefined;
    this.destinationType = undefined;
    this.timezones = [defaultTimezone];
  }

  init = (): void => {
    this._validateWaypoints();
    this._setTimezones();
    this._setPointTypes();
  };

  createRoutes = async (
    roles: Array<Role>,
    georouter: IGeorouter,
  ): Promise<void> => {
    let driverWaypoints: Array<Waypoint> = [];
    let passengerWaypoints: Array<Waypoint> = [];
    const paths: Array<Path> = [];
    if (roles.includes(Role.DRIVER) && roles.includes(Role.PASSENGER)) {
      if (this._points.length == 2) {
        // 2 points => same route for driver and passenger
        const commonPath: Path = {
          key: RouteKey.COMMON,
          points: this._points,
        };
        driverWaypoints = this._createWaypoints(commonPath.points, Role.DRIVER);
        passengerWaypoints = this._createWaypoints(
          commonPath.points,
          Role.PASSENGER,
        );
        paths.push(commonPath);
      } else {
        const driverPath: Path = {
          key: RouteKey.DRIVER,
          points: this._points,
        };
        driverWaypoints = this._createWaypoints(driverPath.points, Role.DRIVER);
        const passengerPath: Path = {
          key: RouteKey.PASSENGER,
          points: [this._points[0], this._points[this._points.length - 1]],
        };
        passengerWaypoints = this._createWaypoints(
          passengerPath.points,
          Role.PASSENGER,
        );
        paths.push(driverPath, passengerPath);
      }
    } else if (roles.includes(Role.DRIVER)) {
      const driverPath: Path = {
        key: RouteKey.DRIVER,
        points: this._points,
      };
      driverWaypoints = this._createWaypoints(driverPath.points, Role.DRIVER);
      paths.push(driverPath);
    } else if (roles.includes(Role.PASSENGER)) {
      const passengerPath: Path = {
        key: RouteKey.PASSENGER,
        points: [this._points[0], this._points[this._points.length - 1]],
      };
      passengerWaypoints = this._createWaypoints(
        passengerPath.points,
        Role.PASSENGER,
      );
      paths.push(passengerPath);
    }
    const routes = await georouter.route(paths, {
      withDistance: false,
      withPoints: true,
      withTime: false,
    });
    if (routes.some((route) => route.key == RouteKey.COMMON)) {
      this.driverRoute = routes.find(
        (route) => route.key == RouteKey.COMMON,
      ).route;
      this.passengerRoute = routes.find(
        (route) => route.key == RouteKey.COMMON,
      ).route;
      this.driverRoute.setWaypoints(driverWaypoints);
      this.passengerRoute.setWaypoints(passengerWaypoints);
    } else {
      if (routes.some((route) => route.key == RouteKey.DRIVER)) {
        this.driverRoute = routes.find(
          (route) => route.key == RouteKey.DRIVER,
        ).route;
        this.driverRoute.setWaypoints(driverWaypoints);
      }
      if (routes.some((route) => route.key == RouteKey.PASSENGER)) {
        this.passengerRoute = routes.find(
          (route) => route.key == RouteKey.PASSENGER,
        ).route;
        this.passengerRoute.setWaypoints(passengerWaypoints);
      }
    }
  };

  _validateWaypoints = (): void => {
    if (this._geographyRequest.waypoints.length < 2) {
      throw new MatcherException(
        MatcherExceptionCode.INVALID_ARGUMENT,
        'At least 2 waypoints are required',
      );
    }
    this._geographyRequest.waypoints.map((point) => {
      if (!this._isValidPoint(point)) {
        throw new MatcherException(
          MatcherExceptionCode.INVALID_ARGUMENT,
          `Waypoint { Lon: ${point.lon}, Lat: ${point.lat} } is not valid`,
        );
      }
      this._points.push(point);
    });
  };

  _setTimezones = (): void => {
    this.timezones = find(
      this._geographyRequest.waypoints[0].lat,
      this._geographyRequest.waypoints[0].lon,
    );
  };

  _setPointTypes = (): void => {
    this.originType =
      this._geographyRequest.waypoints[0].type ?? PointType.OTHER;
    this.destinationType =
      this._geographyRequest.waypoints[
        this._geographyRequest.waypoints.length - 1
      ].type ?? PointType.OTHER;
  };

  _isValidPoint = (point: Point): boolean =>
    this._isValidLongitude(point.lon) && this._isValidLatitude(point.lat);

  _isValidLongitude = (longitude: number): boolean =>
    longitude >= -180 && longitude <= 180;

  _isValidLatitude = (latitude: number): boolean =>
    latitude >= -90 && latitude <= 90;

  _createWaypoints = (points: Array<Point>, role: Role): Array<Waypoint> => {
    return points.map((point, index) => {
      const waypoint = new Waypoint(point);
      if (index == 0) {
        waypoint.addActor(new Actor(this._person, role, Step.START));
      } else if (index == points.length - 1) {
        waypoint.addActor(new Actor(this._person, role, Step.FINISH));
      } else {
        waypoint.addActor(new Actor(this._person, role, Step.INTERMEDIATE));
      }
      return waypoint;
    });
  };
}

export enum RouteKey {
  COMMON = 'common',
  DRIVER = 'driver',
  PASSENGER = 'passenger',
}
